/*
 * bigsort.c
 *
 * Sort big text files on disk
 * Its intention is as a faster way of sorting fidonet user
 * lists when there isn't enough memory.
 *
 * Algorithm is:
 *   work out how much memory is available for chunks
 *   go through the source file:
 *		read in as big a chunk as possible into memory
 *      sort chunk using qsort
 *		write chunk as temporary file with new extension
 *   ; At this point there should be several sorted files
 *   Determine how many file handles are available
 *   while there is more than one remaining file
 *     open as many chunk files as possible and merge them
 *     into a new larger chunk file.
 *   Delete the old file and rename the remaining file.
 */
 
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "fastfile.h"

/*
 * Constants
 */

typedef enum { FALSE, TRUE } BOOL;
 
#if 1
#define MaxRecords 20000		/* Maximum records in one chunk */
#else
#define MaxRecords 4			/* Maximum records in one chunk */
#endif

#define MaxLineLength 256
#define MaxMergeFiles 20		/* Merge up to 8 files at once */

#define WriteBufferSize 0x8000
#define ReadBufferSize  0x8000
#define SpareMemorySize (WriteBufferSize + sizeof(FASTFILE))

/*
 * Global variables
 */
 
char *tmpDir = ".\\";

/*
 * Space used during merge process... it is placed here, because
 * I get stack overflows otherwise!
 */
 
char buffers[MaxMergeFiles][MaxLineLength];

int stringCompare(void *a, void *b)
{
	char *s1 = *(char **)a;
	char *s2 = *(char **)b;
	static int count = 0;
	
	if( (++count % 10000) == 0)
	{
		putc('.', stdout);
		fflush(stdout);
	}
	
	return stricmp(s1, s2);
}


int mergeFiles(int fileCount, char *finalName)
{
	int error = -1;
	while(fileCount > 1)
	{
		FASTFILE *fdOut;
		FASTFILE *fdIn[MaxMergeFiles];
		int inFiles = 0;
		int inCount = 0;
		int firstFile;
		int outFileNumber = 0;
		char destName[FILENAME_MAX];
		char inName[FILENAME_MAX];
		BOOL finalFile = FALSE;
	
		error = -1;
		
		printf("Merging %d files\n", fileCount);
		
		sprintf(destName, "%sbigsort.tmp", tmpDir);
		while(inCount < fileCount)
		{
			error = -1;
			fdOut = fastopen(destName, FF_WRITE, WriteBufferSize);
			if(fdOut)
			{
			
				/*
				 * Open as many files as possible
				 */
				 
				firstFile = inCount;
				while( (inFiles < MaxMergeFiles) && (inCount < fileCount))
				{
					sprintf(inName, "%sbigsort.%03d", tmpDir, inCount);
					fdIn[inFiles] = fastopen(inName, FF_READ, ReadBufferSize);
					if(fdIn[inFiles] == NULL)
						break;
					inFiles++;
					inCount++;
					printf("Reading %s\n", inName);
				}
				
				if(inFiles == 0)
				{
					printf("Unable to open %s\n", inName);
					goto getOut;
				}
				
				/*
				 * If this will result in only 1 file
				 * then get the output going to the
				 * real destination
				 */
				 
				if( (firstFile == 0) && (inCount == fileCount) )
				{
					finalFile = TRUE;
					fastclose(fdOut);
					remove(destName);
					
					/*
					 * It would not be bad idea to make
					 * a backup in case anything goes wrong
					 */
					
					strcpy(destName, finalName);
					fdOut = fastopen(finalName, FF_WRITE, WriteBufferSize);
					if(fdOut == NULL)
						goto getOut;
				}
				
				/*
				 * Merge the files
				 */
				
				{
					BOOL finished[MaxMergeFiles];
					BOOL anyLeft = FALSE;
					
					int count = 0;
					int lineCount = 0;
					
					printf("Merging");
										 
					/*
					 * Get a line from each file
					 */
					 
					while(count < inFiles)
					{
						if(fastgets(buffers[count], MaxLineLength, fdIn[count]) == NULL)
							finished[count] = TRUE;
						else
						{
							finished[count] = FALSE;
							anyLeft = TRUE;
						}
						count++;
					}
					
					while(anyLeft)
					{
						int count = 0;
						int bestBuffer = -1;
						
		
						if( (++lineCount % 1000) == 0)
						{
							putc('.', stdout);
							fflush(stdout);
						}
						
						/*
						 * Find the buffer with the earliest string
						 */
						 
						while(count < inFiles)
						{
							if(!finished[count])
							{
								if(bestBuffer < 0)
									bestBuffer = count;
								else if(stricmp(buffers[count], buffers[bestBuffer]) < 0)
									bestBuffer = count;
							}
							
							count++;
						}
						 
						/*
						 * Write it to the output file
						 */
						 
						fastputs(buffers[bestBuffer], fdOut);
						fastputs("\r\n", fdOut);
						 
						/*
						 * Get a new line
						 */
						
						if(fastgets(buffers[bestBuffer], MaxLineLength, fdIn[bestBuffer]) == NULL)
						{
							finished[bestBuffer] = TRUE;
						 
							/*
							 * Check that there are some files unfinished
							 */
						 
						 	count = 0;
						 	anyLeft = FALSE;
						 	while(count < inFiles)
						 	{
						 		if(!finished[count])
						 		{
						 			anyLeft = TRUE;
						 			break;
						 		}
						 		count++;
						 	}
						 }
					}
					printf("\n");
				}
				
				/*
				 * Close and delete the files
				 */
				 
				 while(inFiles)
				 {
				 	fastclose(fdIn[--inFiles]);
				 	sprintf(inName, "%sbigsort.%03d", tmpDir, inFiles + firstFile);
				 	printf("Deleting %s\n", inName);
				 	remove(inName);
				 }
				 	
				 fastclose(fdOut);
				 
				 /*
				  * rename temporary file
				  */
				 
				 if(!finalFile)
				 {
					 sprintf(inName, "%sbigsort.%03d", tmpDir, outFileNumber++);
					 rename(destName, inName);
					 printf("Merged to %s\n", inName);
				}
				else
					printf("Merged to %s\n", destName);
				  
			}
			else
				printf("Couldn't create %s\n", destName);
				
			error = 0;
		}
		fileCount = outFileNumber;
	}
getOut:
	return error;
}


int sortWriteStrings(char **strings, int howMany, int fileNumber, char *destName)
{
	char fileName[FILENAME_MAX];
	FASTFILE *fdOut;
	int error = -1;
	
	printf("Sorting sequence %d with %d lines",
		fileNumber, howMany);
	
	qsort(strings, howMany, sizeof(*strings), stringCompare);
	printf("\n");
	
	if(destName)
		strcpy(fileName, destName);
	else
		sprintf(fileName, "%sbigsort.%03d", tmpDir, fileNumber);
		
	printf("Writing %s", fileName);
	fflush(stdout);
	fdOut = fastopen(fileName, FF_WRITE, WriteBufferSize);
	if(fdOut)
	{
		while(howMany--)
		{
			if((howMany % 1000) == 0)
			{
				putc('.', stdout);
				fflush(stdout);
			}
			if(fastputs(*strings, fdOut) != FE_OK)
			{
			getOutWrite:
				printf("Error writing %s\n", fileName);
				goto getOut;
			}
			if(fastputs("\r\n", fdOut) != FE_OK)
				goto getOutWrite;
			free(*strings);
			strings++;
		}
		error = 0;
		
	getOut:		
		fastclose(fdOut);
	}
	else
		printf("Couldn't open %s\n", fileName);
		
	printf("\n");
		
	return error;
}

int splitFile(char *fileName)
{
	int error = -1;
	FASTFILE *fdIn;
	
	fdIn = fastopen(fileName, FF_READ, 0x10000L);
	if(fdIn)
	{
		char buffer[MaxLineLength];
		char **linePointers = NULL;
		int lineCount = 0;
		int fileCount = 0;
		void *spareMemory;
	
		printf("Reading %s", fileName);
		
		/*
		 * Reserve some memory for the output buffer
		 */
		 
		spareMemory = malloc(SpareMemorySize);
		if(spareMemory)
		{
			linePointers = calloc(MaxRecords, sizeof(char *));
			if(linePointers)
			{		
				while(fastgets(buffer, MaxLineLength, fdIn))
				{
					char *s;
				
					if((lineCount < MaxRecords) &&
					    ( (s = strdup(buffer)) != NULL) )
					{
						if(buffer[0] != 0)
						{
							linePointers[lineCount++] = s;
							if( (lineCount % 1000) == 0)
							{
								putc('.', stdout);
								fflush(stdout);
							}
						}
					}
					else
					{
						/*
						 * We've run out of memory...
						 * sort it write it and start a new file
						 */
				
						printf("\n");
						free(spareMemory);
						spareMemory = NULL;
						if(sortWriteStrings(linePointers, lineCount, fileCount++, NULL) != 0)
							goto getOut;
						lineCount = 0;
						spareMemory = malloc(SpareMemorySize);
						if(spareMemory == NULL)
							goto getOut;
						
						if(buffer[0])
						{
							s = strdup(buffer);
							if(s)
								linePointers[lineCount++] = s;
							else
								goto getOut;
						}
						printf("Reading");
					}
				}
		
				/*
				 * sort and write the final file
				 */
		 
			 	printf("\n");
			 	free(spareMemory);
			 	spareMemory = NULL;
			 	
			 	{
			 		char *name = NULL;
			 		
			 		if(fileCount == 0)
			 			name = fileName;
					if(sortWriteStrings(linePointers, lineCount, fileCount++, name) != 0)
						goto getOut;
				}
	
				error = 0;
	getOut:				
				free(linePointers);
			}
			if(spareMemory)
				free(spareMemory);
		}
		fastclose(fdIn);
		
		if( (error == 0) && (fileCount > 1) )
			error = mergeFiles(fileCount, fileName);
	}
	else
		printf("Couldn't open %s\n", fileName);
		
	return error;
}

int main(int argc, char *argv[])
{
	char *fileName;			/* Complete path\filename of file to be sorted */
	int error = 0;

	printf("BigSort version 1.1\nCopyright 1993, Steven Green, All rights reserved\n\n");
	
	if(argc != 2)
	{
		printf("\n\nUsage:\n  bigsort filename\n\n");
		return -1;
	}
		
	fileName = argv[1];

	error = splitFile(fileName);
	
	return error;
	

}