/*
 * fastfile.c
 *
 * Some routines to do fast buffered I/O for use with my superfast
 * nodelist processor
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "fastfile.h"

#ifdef __TOS__
 #define _FF_READONLY FO_READ
 #define _FF_WRITEONLY FO_WRITE
 #define open(name,mode)	Fopen(name,mode)
 #define close(fp)			Fclose(fp)
 #define read(fp,ad,len)	Fread(fp,len,ad)
 #define write(fp,ad,len)	Fwrite(fp,len,ad)
 #define creat(name)		Fcreate(name, 0)
#else
 #define _FF_READONLY O_RDONLY
 #define _FF_WRITEONLY (O_WRONLY|O_CREAT)
#endif

FASTFILE *fastfiles = NULL;

FASTFILE *fastopen(char *name, int mode, size_t size)
{
	int fp;
	FASTFILE *newfile;
	
	if(mode & FF_WRITE)
		fp = creat(name);
	else
		fp = open(name, _FF_READONLY);
	if(fp < 0)
		return NULL;
	
	if(size == 0)
		size = (mode & FF_WRITE) ? WRITEBUFFERSIZE : READBUFFERSIZE;

	if( (newfile = calloc(sizeof(FASTFILE) + size, 1)) != NULL)
	{
		newfile->fp = fp;
		newfile->mode = mode;
		newfile->error = FE_OK;
		if(mode & FF_COOKED)
			newfile->flags.cooked = 1;
		newfile->bufsize = size;
		newfile->bufstart = (char *)(newfile+1);
		newfile->bufend = newfile->bufstart + size;
		if(mode & FF_WRITE)
			newfile->pos = newfile->bufstart;
		else
			newfile->pos = newfile->bufend;
		newfile->next = fastfiles;
		fastfiles = newfile;
	}
	else
		close(fp);
		
	return newfile;
}

FE_ERROR fastclose(FASTFILE *file)
{
	FASTFILE *last;
	
	if(file)
	{
		if(!file->error && (file->mode & FF_WRITE))
		{
#ifdef DEBUG
			printf("O");
#endif
			if(file->pos > file->bufstart)
				if(write(file->fp, file->bufstart, file->pos - file->bufstart) < 0)
					file->error = FE_WRITE_ERROR;
		}
		if(close(file->fp) < 0)
			file->error = FE_WRITE_ERROR;

		/* Unlink from list */
		
		last = fastfiles;
		if(last == file)
			fastfiles = file->next;
		else
			while(last)
			{
				if(last->next == file)
				{
					last->next = file->next;
					break;
				}
				last = last->next;
			}

		free(file);
	}
	else
		return FE_NOHANDLE;
	return file->error;
}

char *fastgets(char *buf, int length, FASTFILE *file)
{
	char *dest = buf;
	size_t size;
	
	if(file && !file->error && !file->flags.eof)
	{
		for(;;)
		{
			char *ptr = file->pos;
						
			/* Copy to next cr/lf or end of buffer */
			while(ptr < file->bufend)
			{
				char c = *ptr++;
				
				if( (c  == '\r') || (c == '\n') )
				{
					if(!file->flags.eol)
					{
						file->flags.eol = 1;
					end:
						file->pos = ptr;
						*dest = 0;
						return buf;
					}
				}
				else
				{
					*dest++ = c;
					file->flags.eol = 0;
					if(!--length)
						goto end;
				}
			}
			if(file->flags.finished)
			{
				*dest = 0;
				file->flags.eof = 1;
				
				if(buf[0])
					return buf;
				else
					return NULL;
			}
#ifdef DEBUG
			printf("I");
#endif
			size = read(file->fp, file->bufstart, file->bufsize);
			if(size < 0)
			{
				file->error = FE_READ_ERROR;
				file->flags.eof = 1;
				return NULL;
			}
			file->pos = file->bufstart;
			if(size != file->bufsize)
			{
				file->flags.finished = 1;
				file->bufend = file->bufstart + size;
			}
		}
	}
	else
		return NULL;
}

/*
 * return pointer to next line (in actual buffer!)
 */

char *fastreads(FASTFILE *file)
{
	if(file && !file->error && !file->flags.eof)
	{
		char *retval = file->pos;
		size_t size, size1;

		for(;;)
		{
			char *ptr = file->pos;
						
			/* Search for next cr/lf or end of buffer */
			while(ptr < file->bufend)
			{
				char c = *ptr;
				
				if( (c  == '\r') || (c == '\n') )
				{
					if(file->flags.eol)
					{
						file->flags.eol = 0;
						retval = ++ptr;
					}
					else
					{
						file->flags.eol = 1;
						*ptr++ = 0;
						file->pos = ptr;
						return retval;
					}
				}
				else
				{
					file->flags.eol = 0;
					ptr++;
				}
			}
			if(file->flags.finished)
			{
				*ptr = 0;
				file->flags.eof = 1;
				return retval;
			}
#ifdef DEBUG
			printf("I");
#endif
			size = ptr - retval;
			if(size)
				memcpy(file->bufstart, retval, size);
			retval = file->bufstart;
			file->pos = file->bufstart + size;
			size = file->bufsize - size;
			size1 = read(file->fp, file->pos, size);
			if(size1 < 0)
			{
				file->error = FE_READ_ERROR;
				file->flags.eof = 1;
				return NULL;
			}
			if(size1 != size)
			{
				file->flags.finished = 1;
				file->bufend = file->pos + size1;
			}
		}
	}
	else
		return NULL;
}

FE_ERROR fastputs(char *s, FASTFILE *file)
{
	if(file && !file->error)
	{
		char *ptr = file->pos;
	
		while(*s)
		{
			if(ptr >= file->bufend)
			{
#ifdef DEBUG
				printf("O");
#endif
				if(write(file->fp, file->bufstart, ptr-file->bufstart) < 0)
				{
					file->error = FE_WRITE_ERROR;
					break;
				}
				ptr = file->bufstart;
			}
			*ptr++ = *s++;
		}
		file->pos = ptr;	
		return file->error;
	}
	else
		return FE_WRITE_ERROR;
}

FE_ERROR fastwrite(FASTFILE *file, void *ad, size_t length)
{
	if(file && !file->error)
	{
		if( (file->bufend - file->pos) >= length)
		{
			memcpy(file->pos, ad, length);
			file->pos += length;
			return FE_OK;
		}
		else
		{
#ifdef DEBUG
			printf("O");
#endif
			if(write(file->fp, file->bufstart, file->pos - file->bufstart) < 0)
				file->error = FE_WRITE_ERROR;
			if(write(file->fp, ad, length) < 0)
				file->error = FE_WRITE_ERROR;
		}
		file->pos = file->bufstart;
		return file->error;
	}
	else
		return FE_WRITE_ERROR;
}


int copyfile(char *srcname, char *destname)
{
	FASTFILE *in;
	FASTFILE *out;
	int error = 0;
	
	in = fastopen(srcname, FF_READ|FF_COOKED, 0L);
	if(in != NULL)
	{
		out = fastopen(destname, FF_WRITE|FF_RAW, 0L);
		if(out != NULL)
		{
#if 0
			char buffer[256];
		
			while(fastgets(buffer, 256, in) != NULL)
			{
				fastputs(buffer, out);
				fastputs("\r\n", out);
			}
#else
			char *line;
			
			while( (line = fastreads(in)) != NULL)
			{
				fastputs(line, out);
				fastputs("\r\n", out);
			}
#endif		
			fastclose(out);
		}
		else
			error = -1;
		fastclose(in);
	}
	else
		error =  -1;

	return error;
}

